"""Tests for the modulemd linter."""

import copy
import os
from tempfile import NamedTemporaryFile

from _fedmod.modulemd_linter import LintCheckFailed, ModuleMDLinter

from decorator import decorator

from pytest import fixture, mark, raises

import yaml


testfiles_dir = os.path.join(os.path.dirname(__file__), 'files')
spec_v1_yaml_path = os.path.join(testfiles_dir, 'spec.v1.yaml')
spec_v2_yaml_path = os.path.join(testfiles_dir, 'spec.v2.yaml')

# Load the respective specification YAML files for the different modulemd
# versions. This allows testing the detection of individual defects which are
# introduced in copies of these structures. Do NOT modify these originals!

with open(spec_v1_yaml_path, "r") as f:
    spec_v1_yaml_tree = yaml.load(f)

with open(spec_v2_yaml_path, "r") as f:
    spec_v2_yaml_tree = yaml.load(f)


@decorator
def linting_fails(fn, self, *args, **kwargs):
    """Decorator to check that defective modulemd trees fail."""

    if 'tree' in kwargs:
        tree = kwargs['tree']
    else:
        idx = fn.__code__.co_varnames.index('tree') - 1
        tree = args[idx]

    fn(self, *args, **kwargs)

    with raises(LintCheckFailed):
        self.lint(tree)


class _TestModuleMDLinter(object):
    """Base class for testing different versions of the modulemd format. Should
    contain all tests which are common between the versions."""

    # The parsed spec YAML tree, defined in a child class.
    yaml_tree = None

    @fixture(name='tree')
    @staticmethod
    def fixture_tree(request):
        """Fixture to inject a copy of the parsed YAML tree."""
        self = request.instance
        return self.get_yaml_tree_copy()

    @staticmethod
    def _yaml_tree_remove_infra_only(yaml_tree):
        """Remove YAML elements to be put in by the build infrastructure."""
        try:
            data_tree = yaml_tree['data']
        except KeyError:
            return

        for elem in ('name', 'stream', 'version', 'context', 'arch',
                     'servicelevels', 'eol', 'artifacts'):
            data_tree.pop(elem, None)

        data_tree.get('license', {}).pop('content', None)

    @classmethod
    def get_yaml_tree_copy(cls):
        """Return a cleaned version of the spec YAML structure, without entries
        that aren't supposed to be in the file, but added later by the build
        infrastructure."""

        if not hasattr(cls, '_yaml_tree'):
            # The spec YAML files contain entries that are only supposed to be
            # defined by the build system and not in the original modulemd YAML
            # file. Copy the original structure and remove offending entries.
            cls._yaml_tree = copy.deepcopy(cls.yaml_tree)
            cls._yaml_tree_remove_infra_only(cls._yaml_tree)

        return copy.deepcopy(cls._yaml_tree)

    def _get_linter(self, yaml_tree):
        modulemd_str = yaml.dump(yaml_tree)
        return ModuleMDLinter(modulemd_str=modulemd_str)

    def lint(self, yaml_tree):
        """Run the linter on a (modified) modulemd tree and raise
        LintCheckFailed if anything gets flagged."""
        if self._get_linter(yaml_tree).lint():
            raise LintCheckFailed()


class TestModuleMDGenericLinter(_TestModuleMDLinter):
    """Test generic modulemd structures.

    This class tests generic YAML structures before the version of the format
    is established."""

    yaml_tree = spec_v2_yaml_tree

    def test_non_unicode_document(self):
        """Test that a non-unicode document gets flagged."""
        tree = self.get_yaml_tree_copy()
        tree_bytes = ("äöü\n".encode('iso-8859-15')
                      + yaml.dump(tree).encode('iso-8859-15'))
        with NamedTemporaryFile(mode='wb') as tmpfile:
            tmpfile.write(tree_bytes)
            tmpfile.flush()
            linter = ModuleMDLinter(modulemd_path=tmpfile.name)
            assert linter.lint()

    def test_non_yaml_document(self):
        """Test that a non-YAML document gets flagged."""
        linter = ModuleMDLinter(modulemd_str="not YAML")
        assert linter.lint()

    @linting_fails
    def test_document_missing(self, tree):
        """Test that a document missing the toplevel 'document' tag gets
        flagged."""
        del tree['document']

    @linting_fails
    def test_version_missing(self, tree):
        """Test that a document missing the toplevel 'version' tag gets
        flagged."""
        del tree['version']

    @linting_fails
    def test_wrong_version(self, tree):
        """Test that a document with a wrong 'version' tag gets flagged."""
        tree['version'] = 'giraffe'


class _TestModuleMDVersionLinter(_TestModuleMDLinter):
    """Test modulemd structures common between versions.

    This class contains all tests that can be run against modulemd documents of
    any version."""

    def test_spec_passes(self):
        """Test that known to be good documents pass."""
        self.lint(self.get_yaml_tree_copy())

    @linting_fails
    def test_summary_missing(self, tree):
        """Test that a document without summary gets flagged."""
        del tree['data']['summary']

    @linting_fails
    def test_summary_not_one_sentence(self, tree):
        """Test that a summary with more than one sentence gets flagged."""
        tree['data']['summary'] = "Two. Sentences."

    @linting_fails
    def test_summary_ends_in_period(self, tree):
        """Test that a summary ending in a period gets flagged."""
        tree['data']['summary'] += "."

    @linting_fails
    def test_description_missing(self, tree):
        """Test that a document without description gets flagged."""
        del tree['data']['description']

    @linting_fails
    def test_description_must_end_in_period(self, tree):
        """Test that a description not ending in a period gets flagged."""
        tree['data']['description'] = "Doesn't end in period"

    @linting_fails
    def test_license_missing(self, tree):
        """Test that a missing license block gets flagged."""
        del tree['data']['license']

    @linting_fails
    def test_license_not_dict(self, tree):
        """Test that a non-dict license block gets flagged."""
        tree['data']['license'] = 'yes'

    @linting_fails
    def test_license_subblock_unknown(self, tree):
        """Test that an unknown license subblock gets flagged."""
        tree['data']['license']['humpty'] = ['dumpty']

    @linting_fails
    def test_license_subblock_not_list(self, tree):
        """Test that a license subblock that isn't a list is flagged."""
        tree['data']['license']['module'] = 'yes'

    @linting_fails
    def test_license_module_missing(self, tree):
        """Test that a missing license.module block gets flagged."""
        del tree['data']['license']['module']

    @linting_fails
    def test_license_module_non_scalars(self, tree):
        """Test that a non-scalar license value gets flagged."""
        tree['data']['license']['module'].append({'not': 'a scalar'})

    @mark.parametrize('item', ('name', 'stream', 'version', 'context', 'arch',
                               'eol'))
    @linting_fails
    def test_infra_items_flagged(self, tree, item):
        tree['data'][item] = item

    @linting_fails
    def test_dependencies_missing(self, tree):
        del tree['data']['dependencies']

    @linting_fails
    def test_xmd_not_dict(self, tree):
        tree['data']['xmd'] = ['not a dict']

    @linting_fails
    def test_references_not_dict(self, tree):
        tree['data']['references'] = ['not a dict']

    @linting_fails
    def test_references_unknown_key(self, tree):
        tree['data']['references']['unknown'] = 'foo'

    @linting_fails
    def test_references_non_scalar_value(self, tree):
        tree['data']['references']['documentation'] = ['not scalar']

    @linting_fails
    def test_profiles_not_dict(self, tree):
        tree['data']['profiles'] = ['foo', 'bar', 'baz']

    @linting_fails
    def test_profiles_blocks_not_dict(self, tree):
        tree['data']['profiles']['not dict'] = "not dict"

    @linting_fails
    def test_api_not_dict(self, tree):
        tree['data']['api'] = ('not', 'a', 'dict')

    @linting_fails
    def test_api_blocks_not_list(self, tree):
        tree['data']['api']['rpms'] = 'rpms'

    @linting_fails
    def test_filter_not_dict(self, tree):
        tree['data']['filter'] = ('not', 'a', 'dict')

    @linting_fails
    def test_filter_blocks_not_list(self, tree):
        tree['data']['filter']['rpms'] = {'foo': 'bar'}


class TestModuleMDV1Linter(_TestModuleMDVersionLinter):
    """Test modulemd v1 structures.

    This class contains all tests specific to version 1 of the modulemd format
    and inherits common tests from its base class."""

    yaml_tree = spec_v1_yaml_tree

    @linting_fails
    def test_v2_dependency_tree(self, tree):
        """Test that a version 1 document with a version 2 style 'dependencies'
        block gets flagged."""
        tree['data']['dependencies'] = copy.deepcopy(
            spec_v2_yaml_tree['data']['dependencies'])


class TestModuleMDV2Linter(_TestModuleMDVersionLinter):
    """Test modulemd v2 structures.

    This class contains all tests specific to version 2 of the modulemd format
    and inherits common tests from its base class."""

    yaml_tree = spec_v2_yaml_tree

    @linting_fails
    def test_v1_dependency_tree(self, tree):
        """Test that a version 2 document with a version 1 style 'dependencies'
        block gets flagged."""
        tree['data']['dependencies'] = copy.deepcopy(
            spec_v1_yaml_tree['data']['dependencies'])
